@@ -1,6 +1,6 @@
 /*************************************************************************************************
  * The command line utility of the core API
- *                                                      Copyright (C) 2007-2008 Mikio Hirabayashi
+ *                                                               Copyright (C) 2007-2010 FAL Labs
  * This file is part of Tokyo Dystopia.
  * Tokyo Dystopia is free software; you can redistribute it and/or modify it under the terms of
  * the GNU Lesser General Public License as published by the Free Software Foundation; either
@@ -69,7 +69,7 @@ int main(int argc, char **argv){
   g_progname = argv[0];
   g_dbgfd = -1;
   const char *ebuf = getenv("TCDBGFD");
-  if(ebuf) g_dbgfd = atoi(ebuf);
+  if(ebuf) g_dbgfd = tcatoi(ebuf);
   if(argc < 2) usage();
   int rv = 0;
   if(!strcmp(argv[1], "create")){
@@ -101,10 +101,10 @@ int main(int argc, char **argv){
 
 /* print the usage and exit */
 static void usage(void){
-  fprintf(stderr, "%s: the command line utility of the q-gram database API\n", g_progname);
+  fprintf(stderr, "%s: the command line utility of the core API\n", g_progname);
   fprintf(stderr, "\n");
   fprintf(stderr, "usage:\n");
-  fprintf(stderr, "  %s create [-tl] [-td|-tb] path [ernum [etnum]]\n", g_progname);
+  fprintf(stderr, "  %s create [-tl] [-td|-tb|-tt] path [ernum [etnum]]\n", g_progname);
   fprintf(stderr, "  %s inform [-nl|-nb] path\n", g_progname);
   fprintf(stderr, "  %s put [-nl|-nb] path id text\n", g_progname);
   fprintf(stderr, "  %s out [-nl|-nb] path id\n", g_progname);
@@ -130,19 +130,26 @@ static void printerr(TCIDB *idb){
 
 /* read a line from a file descriptor */
 static char *mygetline(FILE *ifp){
-  char *buf;
-  int c, len, blen;
-  buf = NULL;
-  len = 0;
-  blen = 256;
+  int len = 0;
+  int blen = 1024;
+  char *buf = tcmalloc(blen);
+  bool end = true;
+  int c;
   while((c = fgetc(ifp)) != EOF){
-    if(blen <= len) blen *= 2;
-    buf = tcrealloc(buf, blen + 1);
+    end = false;
+    if(c == '\0') continue;
+    if(blen <= len){
+      blen *= 2;
+      buf = tcrealloc(buf, blen + 1);
+    }
     if(c == '\n' || c == '\r') c = '\0';
     buf[len++] = c;
     if(c == '\0') break;
   }
-  if(!buf) return NULL;
+  if(end){
+    tcfree(buf);
+    return NULL;
+  }
   buf[len] = '\0';
   return buf;
 }
@@ -194,6 +201,8 @@ static int runcreate(int argc, char **argv){
       } else if(!strcmp(argv[i], "-td")){
         opts |= IDBTDEFLATE;
       } else if(!strcmp(argv[i], "-tb")){
+        opts |= IDBTBZIP;
+      } else if(!strcmp(argv[i], "-tt")){
         opts |= IDBTTCBS;
       } else {
         usage();
@@ -374,7 +383,7 @@ static int runsearch(int argc, char **argv){
         smode = IDBSTOKSUF;
       } else if(!strcmp(argv[i], "-max")){
         if(++i >= argc) usage();
-        max = atoi(argv[i]);
+        max = tcatoi(argv[i]);
       } else if(!strcmp(argv[i], "-ph")){
         ph = true;
       } else if(!strcmp(argv[i], "-pv")){
@@ -389,6 +398,7 @@ static int runsearch(int argc, char **argv){
     }
   }
   if(!path || wnum < 1) usage();
+  if(emode == EMCOMP && smode != IDBSSUBSTR) emode = EMISECT;
   if(max < 0) max = INT_MAX;
   int rv = procsearch(path, (const char **)words, wnum, omode, emode, smode, max, ph, pv);
   return rv;
@@ -434,7 +444,7 @@ static int runlist(int argc, char **argv){
         omode |= IDBOLCKNB;
       } else if(!strcmp(argv[i], "-max")){
         if(++i >= argc) usage();
-        max = atoi(argv[i]);
+        max = tcatoi(argv[i]);
       } else if(!strcmp(argv[i], "-pv")){
         pv = true;
       } else {
@@ -463,15 +473,7 @@ static int runimporttsv(int argc, char **argv){
     if(!path && argv[i][0] == '-'){
       if(!strcmp(argv[i], "-ic")){
         if(++i >= argc) usage();
-        char *suffix;
-        icsiz = strtoll(argv[i], &suffix, 10);
-        if(*suffix == 'k' || *suffix == 'K'){
-          icsiz *= 1024LL;
-        } else if(*suffix == 'm' || *suffix == 'M'){
-          icsiz *= 1024LL * 1024LL;
-        } else if(*suffix == 'g' || *suffix == 'G'){
-          icsiz *= 1024LL * 1024LL * 1024LL;
-        }
+        icsiz = tcatoix(argv[i]);
       } else if(!strcmp(argv[i], "-nl")){
         omode |= HDBONOLCK;
       } else if(!strcmp(argv[i], "-nb")){
@@ -683,15 +685,15 @@ static int procsearch(const char *path, const char **words, int wnum, int omode,
     int rnum;
     uint64_t *res;
     switch(emode){
-    case EMISECT:
-      res = tcqdbresisect(rsets, wnum, &rnum);
-      break;
-    case EMDIFF:
-      res = tcqdbresdiff(rsets, wnum, &rnum);
-      break;
-    default:
-      res = tcqdbresunion(rsets, wnum, &rnum);
-      break;
+      case EMISECT:
+        res = tcqdbresisect(rsets, wnum, &rnum);
+        break;
+      case EMDIFF:
+        res = tcqdbresdiff(rsets, wnum, &rnum);
+        break;
+      default:
+        res = tcqdbresunion(rsets, wnum, &rnum);
+        break;
     }
     printresult(idb, res, rnum, max, ph, pv, tctime() - stime);
     tcfree(res);
@@ -829,8 +831,9 @@ static int procimporttsv(const char *path, const char *file, int64_t icsiz, int
 
 /* perform version command */
 static int procversion(void){
-  printf("Tokyo Dystopia version %s (%d:%s)\n", tdversion, _TD_LIBVER, _TD_FORMATVER);
-  printf("Copyright (C) 2007-2008 Mikio Hirabayashi\n");
+  printf("Tokyo Dystopia version %s (%d:%s) for %s\n",
+         tdversion, _TD_LIBVER, _TD_FORMATVER, TDSYSNAME);
+  printf("Copyright (C) 2007-2010 FAL Labs\n");
   return 0;
 }
 
