@@ -1,6 +1,6 @@
 /*************************************************************************************************
  * The core API of Tokyo Dystopia
- *                                                      Copyright (C) 2007-2008 Mikio Hirabayashi
+ *                                                               Copyright (C) 2007-2010 FAL Labs
  * This file is part of Tokyo Dystopia.
  * Tokyo Dystopia is free software; you can redistribute it and/or modify it under the terms of
  * the GNU Lesser General Public License as published by the Free Software Foundation; either
@@ -20,13 +20,13 @@
 #define IDBDIRMODE     00755             // permission of created directories
 #define IDBIOBUFSIZ    65536             // size of an I/O buffer
 #define IDBDEFFWMMAX   2048              // default maximum number forward matching expansion
-#define IDXTXNAME      "text.tch"        // name of the text database
-#define IDXTXMAGIC     0x49              // magic data for identification
+#define IDBTXDBNAME    "dystopia.tch"    // name of the text database
+#define IDBTXDBMAGIC   0x49              // magic data for identification
 
 #define IDBDEFERNUM    1000000           // default expected record number
 #define IDBDEFETNUM    1000000           // default expected token number
-#define IDBDEFIUSIZ    (1024LL*1024*256) // default expected token number
-#define IDBTXBNUMCO    2                 // coefficient of the bucket number of the text database
+#define IDBDEFIUSIZ    (1024LL*1024*512) // default unit size of each index file
+#define IDBTXBNUMCO    2                 // coefficient of the bucket number
 #define IDBTXAPOW      3                 // alignment power of the text database
 #define IDBTXFPOW      10                // free block pool power of the text database
 
@@ -81,7 +81,6 @@ TCIDB *tcidbnew(void){
   idb->etnum = IDBDEFETNUM;
   idb->iusiz = IDBDEFIUSIZ;
   idb->opts = 0;
-  idb->ecode = TCESUCCESS;
   idb->synccb = NULL;
   idb->syncopq = NULL;
   idb->exopts = 0;
@@ -227,7 +226,7 @@ bool tcidbout(TCIDB *idb, int64_t id){
 
 /* Retrieve a record of an indexed database object. */
 char *tcidbget(TCIDB *idb, int64_t id){
- assert(idb && id > 0);
+  assert(idb && id > 0);
   if(!tcidblockmethod(idb, false)) return false;
   if(!idb->path){
     tchdbsetecode(idb->txdb, TCEINVALID, __FILE__, __LINE__, __func__);
@@ -340,7 +339,17 @@ uint64_t *tcidbsearch2(TCIDB *idb, const char *expr, int *np){
   }
   tcfree(nexpr);
   int tnum = tclistnum(terms);
-  QDBRSET *rsets = tcmalloc((tnum + 1) * sizeof(*rsets));
+  if(tnum < 1){
+    tclistdel(terms);
+    *np = 0;
+    return tcmalloc(1);
+  }
+  if(tnum == 1){
+    uint64_t *res = tcidbsearchtoken(idb, tclistval2(terms, 0), np);
+    tclistdel(terms);
+    return res;
+  }
+  QDBRSET *rsets = tcmalloc(tnum * sizeof(*rsets));
   int rsnum = 0;
   bool sign = true;
   int ti = 0;
@@ -674,7 +683,7 @@ static bool tcidblockmethod(TCIDB *idb, bool wr){
 
 
 /* Unlock a method of the indexed database object.
-   `bdb' specifies the indexed database object.
+   `idb' specifies the indexed database object.
    If successful, the return value is true, else, it is false. */
 static bool tcidbunlockmethod(TCIDB *idb){
   assert(idb);
@@ -695,10 +704,24 @@ static bool tcidbunlockmethod(TCIDB *idb){
 static bool tcidbsynccb(int total, int current, const char *msg, TCIDB *idb){
   bool rv = idb->synccb ? idb->synccb(total, current, msg, idb->syncopq) : true;
   if((total|current) == 0 && !strcmp(msg, QDBSYNCMSGL) &&
-     tcqdbfsiz(idb->idxs[idb->cnum]) >= idb->iusiz){
+     tcqdbfsiz(idb->idxs[idb->cnum]) >= idb->iusiz && idb->inum > 0){
+    TCQDB **idxs = idb->idxs;
     if(idb->synccb && !idb->synccb(total, current, "to be cycled", idb->syncopq)) rv = false;
-    idb->cnum++;
-    if(idb->cnum >= IDBQDBMAX) idb->cnum = 0;
+    if(!tcqdbcacheclear(idxs[idb->cnum])){
+      tchdbsetecode(idb->txdb, tcqdbecode(idxs[idb->cnum]), __FILE__, __LINE__, __func__);
+      rv = false;
+    }
+    int inum = idb->inum;
+    idb->cnum = 0;
+    uint64_t min = UINT64_MAX;
+    for(int i = 0; i < inum; i++){
+      uint64_t fsiz = tcqdbfsiz(idxs[i]);
+      if(fsiz < min){
+        idb->cnum = i;
+        min = fsiz;
+      }
+    }
+    if(min > idb->iusiz && inum < IDBQDBMAX) idb->cnum = inum;
   }
   return rv;
 }
@@ -717,21 +740,21 @@ static bool tcidbopenimpl(TCIDB *idb, const char *path, int omode){
       if(mkdir(path, IDBDIRMODE) == -1 && errno != EEXIST){
         int ecode = TCEMKDIR;
         switch(errno){
-        case EACCES: ecode = TCENOPERM; break;
-        case ENOENT: ecode = TCENOFILE; break;
+          case EACCES: ecode = TCENOPERM; break;
+          case ENOENT: ecode = TCENOFILE; break;
         }
         tchdbsetecode(idb->txdb, ecode, __FILE__, __LINE__, __func__);
         return false;
       }
     }
     if(omode & IDBOTRUNC){
-      sprintf(pbuf, "%s%c%s", path, MYPATHCHR, IDXTXNAME);
+      sprintf(pbuf, "%s%c%s", path, MYPATHCHR, IDBTXDBNAME);
       if(unlink(pbuf) == -1 && errno != ENOENT){
         tchdbsetecode(idb->txdb, TCEUNLINK, __FILE__, __LINE__, __func__);
         return false;
       }
       for(int i = 0; i < IDBQDBMAX; i++){
-        sprintf(pbuf, "%s%c%04d", path, MYPATHCHR, i);
+        sprintf(pbuf, "%s%c%04d", path, MYPATHCHR, i + 1);
         if(unlink(pbuf) == -1 && errno != ENOENT){
           tchdbsetecode(idb->txdb, TCEUNLINK, __FILE__, __LINE__, __func__);
           return false;
@@ -743,8 +766,8 @@ static bool tcidbopenimpl(TCIDB *idb, const char *path, int omode){
   if(stat(path, &sbuf) == -1){
     int ecode = TCEOPEN;
     switch(errno){
-    case EACCES: ecode = TCENOPERM; break;
-    case ENOENT: ecode = TCENOFILE; break;
+      case EACCES: ecode = TCENOPERM; break;
+      case ENOENT: ecode = TCENOFILE; break;
     }
     tchdbsetecode(idb->txdb, ecode, __FILE__, __LINE__, __func__);
     return false;
@@ -760,6 +783,7 @@ static bool tcidbopenimpl(TCIDB *idb, const char *path, int omode){
   int qomode = QDBOREADER;
   uint8_t qopts = 0;
   int64_t etnum = idb->etnum;
+  int64_t iusiz = idb->iusiz;
   if(omode & IDBOWRITER){
     homode = HDBOWRITER;
     qomode = QDBOWRITER;
@@ -777,6 +801,7 @@ static bool tcidbopenimpl(TCIDB *idb, const char *path, int omode){
       qopts |= QDBTLARGE;
     }
     if(idb->opts & IDBTDEFLATE) qopts |= QDBTDEFLATE;
+    if(idb->opts & IDBTBZIP) qopts |= QDBTBZIP;
     if(idb->opts & IDBTTCBS){
       hopts |= HDBTTCBS;
       qopts |= QDBTTCBS;
@@ -795,19 +820,25 @@ static bool tcidbopenimpl(TCIDB *idb, const char *path, int omode){
     homode |= HDBOLCKNB;
     qomode |= QDBOLCKNB;
   }
-  sprintf(pbuf, "%s%c%s", path, MYPATHCHR, IDXTXNAME);
+  sprintf(pbuf, "%s%c%s", path, MYPATHCHR, IDBTXDBNAME);
   if(!tchdbopen(txdb, pbuf, homode)) return false;
   char *txopq = tchdbopaque(txdb);
   uint8_t magic = *(uint8_t *)txopq;
   if(magic == 0 && (omode & IDBOWRITER)){
-    *(uint8_t *)txopq = IDXTXMAGIC;
+    *(uint8_t *)txopq = IDBTXDBMAGIC;
     *(uint8_t *)(txopq + sizeof(magic) + sizeof(uint8_t)) = qopts;
     uint64_t llnum = TDHTOILL(etnum);
     memcpy(txopq + sizeof(magic) + sizeof(uint8_t) + sizeof(qopts), &llnum, sizeof(llnum));
+    llnum = TDHTOILL(iusiz);
+    memcpy(txopq + sizeof(magic) + sizeof(uint8_t) + sizeof(qopts) + sizeof(llnum),
+           &llnum, sizeof(llnum));
   } else {
     qopts = *(uint8_t *)(txopq + sizeof(magic) + sizeof(uint8_t));
     memcpy(&etnum, txopq + sizeof(magic) + sizeof(uint8_t) + sizeof(qopts), sizeof(etnum));
     etnum = TDITOHLL(etnum);
+    memcpy(&iusiz, txopq + sizeof(magic) + sizeof(uint8_t) + sizeof(qopts) + sizeof(etnum),
+           sizeof(iusiz));
+    iusiz = TDITOHLL(iusiz);
   }
   if(omode & IDBOWRITER){
     for(int i = 0; i < IDBQDBMAX; i++){
@@ -820,6 +851,7 @@ static bool tcidbopenimpl(TCIDB *idb, const char *path, int omode){
   idb->opts = 0;
   if(qopts & QDBTLARGE) idb->opts |= QDBTLARGE;
   if(qopts & QDBTDEFLATE) idb->opts |= QDBTDEFLATE;
+  if(qopts & QDBTBZIP) idb->opts |= QDBTBZIP;
   if(qopts & QDBTTCBS) idb->opts |= IDBTTCBS;
   uint8_t inum;
   memcpy(&inum, txopq + sizeof(magic), sizeof(inum));
@@ -851,6 +883,7 @@ static bool tcidbopenimpl(TCIDB *idb, const char *path, int omode){
   idb->wmode = omode & IDBOWRITER;
   idb->qopts = qopts;
   idb->qomode = qomode;
+  idb->iusiz = iusiz;
   return true;
 }
 
@@ -868,6 +901,7 @@ static bool tcidbcloseimpl(TCIDB *idb){
     char *txopq = tchdbopaque(txdb);
     *(uint8_t *)(txopq + sizeof(uint8_t)) = inum;
   }
+  idb->inum = 0;
   for(int i = 0; i < inum; i++){
     if(!tcqdbclose(idxs[i])){
       tchdbsetecode(txdb, tcqdbecode(idxs[i]), __FILE__, __LINE__, __func__);
@@ -910,7 +944,7 @@ static bool tcidbputimpl(TCIDB *idb, int64_t id, const char *text){
   char stack[IDBIOBUFSIZ];
   int vsiz = tchdbget3(txdb, kbuf, ksiz, stack, IDBIOBUFSIZ);
   uint8_t ocnum;
-  if(vsiz >= sizeof(ocnum)){
+  if(vsiz >= (int)sizeof(ocnum)){
     ocnum = ((uint8_t *)stack)[vsiz-1];
     if(ocnum >= IDBQDBMAX){
       tchdbsetecode(txdb, TCEMISC, __FILE__, __LINE__, __func__);
@@ -920,7 +954,7 @@ static bool tcidbputimpl(TCIDB *idb, int64_t id, const char *text){
     if(vsiz >= IDBIOBUFSIZ){
       char *vbuf = tchdbget(txdb, kbuf, ksiz, &vsiz);
       if(vbuf){
-        if(vsiz < sizeof(ocnum)){
+        if(vsiz < (int)sizeof(ocnum)){
           tchdbsetecode(txdb, TCEMISC, __FILE__, __LINE__, __func__);
           tcfree(vbuf);
           return false;
@@ -981,7 +1015,7 @@ static bool tcidboutimpl(TCIDB *idb, int64_t id){
   char stack[IDBIOBUFSIZ];
   int vsiz = tchdbget3(txdb, kbuf, ksiz, stack, IDBIOBUFSIZ);
   uint8_t ocnum;
-  if(vsiz >= sizeof(ocnum)){
+  if(vsiz >= (int)sizeof(ocnum)){
     ocnum = ((uint8_t *)stack)[vsiz-1];
     if(ocnum >= IDBQDBMAX){
       tchdbsetecode(txdb, TCEMISC, __FILE__, __LINE__, __func__);
@@ -991,7 +1025,7 @@ static bool tcidboutimpl(TCIDB *idb, int64_t id){
     if(vsiz >= IDBIOBUFSIZ){
       char *vbuf = tchdbget(txdb, kbuf, ksiz, &vsiz);
       if(vbuf){
-        if(vsiz < sizeof(ocnum)){
+        if(vsiz < (int)sizeof(ocnum)){
           tchdbsetecode(txdb, TCEMISC, __FILE__, __LINE__, __func__);
           tcfree(vbuf);
           return false;
@@ -1037,7 +1071,7 @@ static char *tcidbgetimpl(TCIDB *idb, int64_t id){
   int vsiz;
   char *vbuf = tchdbget(idb->txdb, kbuf, ksiz, &vsiz);
   if(!vbuf) return NULL;
-  if(vsiz < sizeof(uint8_t)){
+  if(vsiz < (int)sizeof(uint8_t)){
     tcfree(vbuf);
     tchdbsetecode(idb->txdb, TCEMISC, __FILE__, __LINE__, __func__);
     return false;
@@ -1048,7 +1082,7 @@ static char *tcidbgetimpl(TCIDB *idb, int64_t id){
 
 
 /* Search an indexed database.
-   `qdb' specifies the indexed database object.
+   `idb' specifies the indexed database object.
    `word' specifies the string of the word to be matched to.
    `smode' specifies the matching mode.
    `np' specifies the pointer to the variable into which the number of elements of the return
@@ -1212,17 +1246,17 @@ static bool tcidbcopyimpl(TCIDB *idb, const char *path){
   if(mkdir(path, IDBDIRMODE) == -1 && errno != EEXIST){
     int ecode = TCEMKDIR;
     switch(errno){
-    case EACCES: ecode = TCENOPERM; break;
-    case ENOENT: ecode = TCENOFILE; break;
+      case EACCES: ecode = TCENOPERM; break;
+      case ENOENT: ecode = TCENOFILE; break;
     }
     tchdbsetecode(txdb, ecode, __FILE__, __LINE__, __func__);
     return false;
   }
   char pbuf[strlen(path)+TDNUMBUFSIZ];
-  sprintf(pbuf, "%s%c%s", path, MYPATHCHR, IDXTXNAME);
+  sprintf(pbuf, "%s%c%s", path, MYPATHCHR, IDBTXDBNAME);
   if(!tchdbcopy(txdb, pbuf)) err = true;
   for(int i = 0; i < inum; i++){
-    sprintf(pbuf, "%s%c%04d", path, MYPATHCHR, i);
+    sprintf(pbuf, "%s%c%04d", path, MYPATHCHR, i + 1);
     if(!tcqdbcopy(idxs[i], pbuf)){
       tchdbsetecode(txdb, tcqdbecode(idxs[i]), __FILE__, __LINE__, __func__);
       err = true;
