@@ -1,6 +1,6 @@
 /*************************************************************************************************
  * The q-gram database API of Tokyo Dystopia
- *                                                      Copyright (C) 2007-2008 Mikio Hirabayashi
+ *                                                               Copyright (C) 2007-2010 FAL Labs
  * This file is part of Tokyo Dystopia.
  * Tokyo Dystopia is free software; you can redistribute it and/or modify it under the terms of
  * the GNU Lesser General Public License as published by the Free Software Foundation; either
@@ -17,13 +17,16 @@
 #include "tcqdb.h"
 #include "myconf.h"
 
+#define QDBMAGICDATA   "[q-gram]"        // magic data for identification
 #define QDBIOBUFSIZ    65536             // size of an I/O buffer
 #define QDBMAXWORDLEN  1024              // maximum length of each search word
 #define QDBOCRUNIT     1024              // unit number of occurrence allocation
 #define QDBCCBNUM      1048573           // bucket number of the token cache
 #define QDBCCDEFICSIZ  (1024LL*1024*128) // default capacity of the token cache
+#define QDBZMMINSIZ    131072            // minimum memory size to use nullified region
 #define QDBDIDSBNUM    262139            // bucket number of the deleted ID set
 #define QDBDTKNBNUM    262139            // bucket number of the deleted token map
+#define QDBBITMAPNUM   524287            // number of elements of search bitmap
 #define QDBDEFFWMMAX   2048              // default maximum number forward matching expansion
 #define QDBHJBNUMCO    4                 // coefficient of the bucket number for hash join
 
@@ -39,7 +42,9 @@
 
 typedef struct {                         // type of structure for a record occurrence
   uint64_t id;                           // ID number
-  uint32_t off;                          // offset from the first token
+  int32_t off;                           // offset from the first token
+  uint16_t seq;                          // sequence number
+  uint16_t hash;                         // hash value for counting sort
 } QDBOCR;
 
 
@@ -51,7 +56,7 @@ static bool tcqdbcloseimpl(TCQDB *qdb);
 static bool tcqdbputimpl(TCQDB *qdb, int64_t id, const char *text);
 static bool tcqdboutimpl(TCQDB *qdb, int64_t id, const char *text);
 static uint64_t *tcqdbsearchimpl(TCQDB *qdb, const char *word, int smode, int *np);
-static int tccmptokens(const TCLISTDATUM *a, const TCLISTDATUM *b);
+static int tccmptokens(const uint16_t **a, const uint16_t **b);
 static int tccmpocrs(QDBOCR *a, QDBOCR *b);
 static int tccmpuint64(const uint64_t *a, const uint64_t *b);
 
@@ -68,7 +73,7 @@ const char *tdversion = _TD_VERSION;
 
 /* Get the message string corresponding to an error code. */
 const char *tcqdberrmsg(int ecode){
-  return tchdberrmsg(ecode);
+  return tcbdberrmsg(ecode);
 }
 
 
@@ -387,30 +392,29 @@ bool tcqdbmemsync(TCQDB *qdb, int level){
       tcbdbsetecode(qdb->idx, TCEMISC, __FILE__, __LINE__, __func__);
       return false;
     }
-    TCLIST *keys = tcmapkeys(cc);
-    int kn = TCLISTNUM(keys);
+    int kn;
+    const char **keys = tcmapkeys2(cc, &kn);
     if(synccb && !synccb(kn, 0, "sorting tokens", syncopq)){
       tcbdbsetecode(qdb->idx, TCEMISC, __FILE__, __LINE__, __func__);
-      tclistdel(keys);
+      tcfree(keys);
       return false;
     }
-    tclistsortex(keys, tccmptokens);
+    qsort(keys, kn, sizeof(*keys), (int(*)(const void *, const void *))tccmptokens);
     char token[TDNUMBUFSIZ*2];
     for(int i = 0; i < kn; i++){
       if(synccb && !synccb(kn, i + 1, "storing tokens", syncopq)){
         tcbdbsetecode(qdb->idx, TCEMISC, __FILE__, __LINE__, __func__);
-        tclistdel(keys);
+        tcfree(keys);
         return false;
       }
-      int asiz;
-      const uint16_t *ary = tclistval(keys, i, &asiz);
+      const uint16_t *ary = (uint16_t *)keys[i];
       tcstrucstoutf(ary, 2, token);
       int tlen = strlen(token);
       int vsiz;
-      const char *vbuf = tcmapget(cc, ary, asiz, &vsiz);
+      const char *vbuf = tcmapget(cc, ary, 2 * sizeof(*ary), &vsiz);
       if(!tcbdbputcat(idx, token, tlen, vbuf, vsiz)) err = true;
     }
-    tclistdel(keys);
+    tcfree(keys);
     tcmapclear(cc);
   }
   TCMAP *dtokens = qdb->dtokens;
@@ -420,23 +424,22 @@ bool tcqdbmemsync(TCQDB *qdb, int level){
       tcbdbsetecode(qdb->idx, TCEMISC, __FILE__, __LINE__, __func__);
       return false;
     }
-    TCLIST *keys = tcmapkeys(dtokens);
-    int kn = TCLISTNUM(keys);
+    int kn;
+    const char **keys = tcmapkeys2(dtokens, &kn);
     if(synccb && !synccb(kn, 0, "sorting deleted tokens", syncopq)){
       tcbdbsetecode(qdb->idx, TCEMISC, __FILE__, __LINE__, __func__);
-      tclistdel(keys);
+      tcfree(keys);
       return false;
     }
-    tclistsortex(keys, tccmptokens);
+    qsort(keys, kn, sizeof(*keys), (int(*)(const void *, const void *))tccmptokens);
     char token[TDNUMBUFSIZ*2];
     for(int i = 0; i < kn; i++){
       if(synccb && !synccb(kn, i + 1, "storing deleted tokens", syncopq)){
         tcbdbsetecode(qdb->idx, TCEMISC, __FILE__, __LINE__, __func__);
-        tclistdel(keys);
+        tcfree(keys);
         return false;
       }
-      int asiz;
-      const uint16_t *ary = tclistval(keys, i, &asiz);
+      const uint16_t *ary = (uint16_t *)keys[i];
       tcstrucstoutf(ary, 2, token);
       int tlen = strlen(token);
       int vsiz;
@@ -472,7 +475,7 @@ bool tcqdbmemsync(TCQDB *qdb, int level){
       }
       tcfree(nbuf);
     }
-    tclistdel(keys);
+    tcfree(keys);
     tcmapclear(dtokens);
     tcidsetclear(dids);
   }
@@ -491,6 +494,17 @@ bool tcqdbmemsync(TCQDB *qdb, int level){
 }
 
 
+/* Clear the cache of a q-gram database object. */
+bool tcqdbcacheclear(TCQDB *qdb){
+  assert(qdb);
+  if(!qdb->open){
+    tcbdbsetecode(qdb->idx, TCEINVALID, __FILE__, __LINE__, __func__);
+    return false;
+  }
+  return tcbdbcacheclear(qdb->idx);
+}
+
+
 /* Get the inode number of the database file of a q-gram database object. */
 uint64_t tcqdbinode(TCQDB *qdb){
   assert(qdb);
@@ -512,6 +526,21 @@ uint8_t tcqdbopts(TCQDB *qdb){
 }
 
 
+/* Get the maximum number of forward matching expansion of a q-gram database object. */
+uint32_t tcqdbfwmmax(TCQDB *qdb){
+  assert(qdb);
+  return qdb->fwmmax;
+}
+
+
+/* Get the number of records in the cache of a q-gram database object. */
+uint32_t tcqdbcnum(TCQDB *qdb){
+  assert(qdb);
+  if(!qdb->cc) return 0;
+  return tcmaprnum(qdb->cc);
+}
+
+
 /* Set the callback function for sync progression of a q-gram database object. */
 void tcqdbsetsynccb(TCQDB *qdb, bool (*cb)(int, int, const char *, void *), void *opq){
   assert(qdb);
@@ -981,7 +1010,13 @@ void tctextnormalize(char *text, int opts){
         ary[nnum++] = 0x007c;
       } else if(c >= 0xff21 && c <= 0xff3a){
         // fullwidth alphabets
-        ary[nnum++] = c - 0xfee0;
+        if(lowmode){
+          c -= 0xfee0;
+          if(c >= 0x0041 && c <= 0x005a) c += 0x20;
+          ary[nnum++] = c;
+        } else {
+          ary[nnum++] = c - 0xfee0;
+        }
       } else if(c >= 0xff41 && c <= 0xff5a){
         // fullwidth small alphabets
         ary[nnum++] = c - 0xfee0;
@@ -1088,8 +1123,8 @@ void tctextnormalize(char *text, int opts){
       ary[nnum++] = c;
     }
   }
-  if(ary != stack) tcfree(ary);
   tcstrucstoutf(ary, nnum, text);
+  if(ary != stack) tcfree(ary);
   if(spcmode) tcstrsqzspc(text);
 }
 
@@ -1098,7 +1133,12 @@ void tctextnormalize(char *text, int opts){
 TCIDSET *tcidsetnew(uint32_t bnum){
   if(bnum < 1) bnum = 1;
   TCIDSET *idset = tcmalloc(sizeof(*idset));
-  uint64_t *buckets = tccalloc(bnum, sizeof(*buckets));
+  uint64_t *buckets;
+  if(bnum >= QDBZMMINSIZ / sizeof(*buckets)){
+    buckets = tczeromap(bnum * sizeof(*buckets));
+  } else {
+    buckets = tccalloc(bnum, sizeof(*buckets));
+  }
   idset->buckets = buckets;
   idset->bnum = bnum;
   idset->trails = tcmapnew2((bnum >> 2) + 1);
@@ -1110,7 +1150,11 @@ TCIDSET *tcidsetnew(uint32_t bnum){
 void tcidsetdel(TCIDSET *idset){
   assert(idset);
   tcmapdel(idset->trails);
-  tcfree(idset->buckets);
+  if(idset->bnum >= QDBZMMINSIZ / sizeof(idset->buckets[0])){
+    tczerounmap(idset->buckets);
+  } else {
+    tcfree(idset->buckets);
+  }
   tcfree(idset);
 }
 
@@ -1204,6 +1248,7 @@ static bool tcqdbopenimpl(TCQDB *qdb, const char *path, int omode){
     int bopts = 0;
     if(qdb->opts & QDBTLARGE) bopts |= BDBTLARGE;
     if(qdb->opts & QDBTDEFLATE) bopts |= BDBTDEFLATE;
+    if(qdb->opts & QDBTBZIP) bopts |= BDBTBZIP;
     if(qdb->opts & QDBTTCBS) bopts |= BDBTTCBS;
     if(!tcbdbtune(qdb->idx, QDBLMEMB, QDBNMEMB, bnum, QDBAPOW, QDBFPOW, bopts)) return false;
     if(!tcbdbsetlsmax(qdb->idx, QDBLSMAX)) return false;
@@ -1217,6 +1262,14 @@ static bool tcqdbopenimpl(TCQDB *qdb, const char *path, int omode){
   if(omode & QDBONOLCK) bomode |= BDBONOLCK;
   if(omode & QDBOLCKNB) bomode |= BDBOLCKNB;
   if(!tcbdbopen(qdb->idx, path, bomode)) return false;
+  if((omode & QDBOWRITER) && tcbdbrnum(qdb->idx) < 1){
+    memcpy(tcbdbopaque(qdb->idx), QDBMAGICDATA, strlen(QDBMAGICDATA));
+  } else if(!(omode & QDBONOLCK) &&
+            memcmp(tcbdbopaque(qdb->idx), QDBMAGICDATA, strlen(QDBMAGICDATA))){
+    tcbdbclose(qdb->idx);
+    tcbdbsetecode(qdb->idx, TCEMETA, __FILE__, __LINE__, __func__);
+    return 0;
+  }
   if(omode & QDBOWRITER){
     qdb->cc = tcmapnew2(QDBCCBNUM);
     qdb->dtokens = tcmapnew2(QDBDTKNBNUM);
@@ -1315,256 +1368,251 @@ static bool tcqdboutimpl(TCQDB *qdb, int64_t id, const char *text){
    records. */
 static uint64_t *tcqdbsearchimpl(TCQDB *qdb, const char *word, int smode, int *np){
   assert(qdb && word && np);
-  int wlen = strlen(word);
-  if(wlen > QDBMAXWORDLEN){
-    tcbdbsetecode(qdb->idx, TCEINVALID, __FILE__, __LINE__, __func__);
-    return NULL;
-  }
-  TCBDB *bdb = qdb->idx;
-  uint16_t ary[wlen+1];
+  TCBDB *idx = qdb->idx;
+  uint64_t *res = NULL;
+  int wsiz = strlen(word);
+  uint16_t *ary = tcmalloc(sizeof(*ary) * (wsiz + 2));
   int anum;
   tcstrutftoucs(word, ary, &anum);
-  if(anum < 1){
-    *np = 0;
-    return tcmalloc(1);
+  for(int i = 0; i < 2; i++){
+    ary[anum+i] = 0;
   }
-  ary[anum] = 0x0000;
-  char token[TDNUMBUFSIZ*2];
-  QDBOCR *ocrs = tcmalloc(QDBOCRUNIT * sizeof(*ocrs));
-  int onum = 0;
-  int oanum = QDBOCRUNIT;
-  if(anum == 1){
-    tcstrucstoutf(ary, 2, token);
-    int tlen = strlen(token);
-    TCLIST *keys;
-    if(smode == QDBSSUFFIX || smode == QDBSFULL){
-      keys = tclistnew();
-      tclistpush(keys, token, tlen);
-    } else {
-      keys = tcbdbfwmkeys(bdb, token, tlen, qdb->fwmmax);
-    }
-    int kn = tclistnum(keys);
-    for(int i = 0; i < kn; i++){
-      int ksiz;
-      const char *kbuf = tclistval(keys, i, &ksiz);
-      int vsiz;
-      const char *vbuf = tcbdbget3(bdb, kbuf, ksiz, &vsiz);
-      if(!vbuf) continue;
-      if(smode == QDBSPREFIX || smode == QDBSFULL){
-        while(vsiz > 1){
+  if(anum >= 2){
+    QDBOCR *ocrs = tcmalloc(QDBOCRUNIT * sizeof(*ocrs));
+    int onum = 0;
+    int oanum = QDBOCRUNIT;
+    int obase = 0;
+    TCBITMAP *pkmap = TCBITMAPNEW(QDBBITMAPNUM);
+    char token[2*3+1];
+    uint16_t seq = 0;
+    for(int i = 0; i < anum; i += 2){
+      obase = onum;
+      int diff = anum - i - 2;
+      if(diff < 0){
+        i += diff;
+        diff = -diff;
+      } else {
+        diff = 0;
+      }
+      tcstrucstoutf(ary + i, 2, token);
+      int tsiz = strlen(token);
+      int csiz;
+      const char *cbuf = tcbdbget3(idx, token, tsiz, &csiz);
+      if(cbuf){
+        while(csiz > 0){
+          int64_t id = 0;
           int step;
-          uint64_t id;
-          TDREADVNUMBUF64(vbuf, id, step);
-          vbuf += step;
-          vsiz -= step;
-          if(vsiz > 0){
-            uint32_t off;
-            TDREADVNUMBUF(vbuf, off, step);
-            if(off == 0){
-              vbuf += step;
-              vsiz -= step;
+          TDREADVNUMBUF64(cbuf, id, step);
+          cbuf += step;
+          csiz -= step;
+          if(csiz > 0){
+            int off;
+            TDREADVNUMBUF(cbuf, off, step);
+            cbuf += step;
+            csiz -= step;
+            off += diff;
+            uint32_t hash = id % QDBBITMAPNUM;
+            if(i == 0 || TCBITMAPCHECK(pkmap, hash)){
               if(onum >= oanum){
                 oanum *= 2;
                 ocrs = tcrealloc(ocrs, oanum * sizeof(*ocrs));
               }
-              ocrs[onum].off = off;
-              ocrs[onum].id = id;
+              QDBOCR *ocr = ocrs + onum;
+              ocr->id = id;
+              ocr->off = off;
+              ocr->seq = seq;
+              ocr->hash = hash;
+              onum++;
+              if(i == 0) TCBITMAPON(pkmap, hash);
+            }
+          }
+        }
+        if(i == 0 && (smode == QDBSPREFIX || smode == QDBSFULL)){
+          int nnum = 0;
+          for(int j = 0; j < onum; j++){
+            if(ocrs[j].off == 0) ocrs[nnum++] = ocrs[j];
+          }
+          onum = nnum;
+        }
+      }
+      seq++;
+      if(onum <= obase){
+        onum = 0;
+        break;
+      }
+    }
+    if(smode == QDBSSUFFIX || smode == QDBSFULL){
+      obase = onum;
+      int diff = anum % 2 + 1;
+      tcstrucstoutf(ary + (anum - 1), 2, token);
+      int tsiz = strlen(token);
+      int csiz;
+      const char *cbuf = tcbdbget3(idx, token, tsiz, &csiz);
+      if(cbuf){
+        while(csiz > 0){
+          int64_t id = 0;
+          int step;
+          TDREADVNUMBUF64(cbuf, id, step);
+          cbuf += step;
+          csiz -= step;
+          if(csiz > 0){
+            int off;
+            TDREADVNUMBUF(cbuf, off, step);
+            cbuf += step;
+            csiz -= step;
+            off += diff;
+            uint32_t hash = id % QDBBITMAPNUM;
+            if(TCBITMAPCHECK(pkmap, hash)){
+              if(onum >= oanum){
+                oanum *= 2;
+                ocrs = tcrealloc(ocrs, oanum * sizeof(*ocrs));
+              }
+              QDBOCR *ocr = ocrs + onum;
+              ocr->id = id;
+              ocr->off = off;
+              ocr->seq = seq;
+              ocr->hash = hash;
               onum++;
             }
           }
         }
-      } else {
-        while(vsiz > 1){
-          int step;
-          uint64_t id;
-          TDREADVNUMBUF64(vbuf, id, step);
-          vbuf += step;
-          vsiz -= step;
-          if(vsiz > 0){
-            uint32_t off;
-            TDREADVNUMBUF(vbuf, off, step);
-            vbuf += step;
-            vsiz -= step;
-            if(onum >= oanum){
-              oanum *= 2;
-              ocrs = tcrealloc(ocrs, oanum * sizeof(*ocrs));
-            }
-            ocrs[onum].off = off;
-            ocrs[onum].id = id;
-            onum++;
-          }
+      }
+      seq++;
+      if(onum <= obase) onum = 0;
+    }
+    TCBITMAPDEL(pkmap);
+    if(seq > 1){
+      if(onum > UINT16_MAX){
+        int flnum = onum * 16 + 1;
+        TCBITMAP *flmap = TCBITMAPNEW(flnum);
+        for(int i = obase; i < onum; i++){
+          QDBOCR *ocr = ocrs + i;
+          uint32_t hash = (((uint32_t)ocr->off << 16) | ocr->hash) % flnum;
+          TCBITMAPON(flmap, hash);
+        }
+        int wi = 0;
+        for(int i = 0; i < obase; i++){
+          QDBOCR *ocr = ocrs + i;
+          int rem = (seq - ocr->seq - 1) * 2;
+          uint32_t hash = (((uint32_t)(ocr->off + rem) << 16) | ocr->hash) % flnum;
+          if(TCBITMAPCHECK(flmap, hash)) ocrs[wi++] = *ocr;
+        }
+        for(int i = obase; i < onum; i++){
+          ocrs[wi++] = ocrs[i];
+        }
+        onum = wi;
+        TCBITMAPDEL(flmap);
+      }
+      if(onum > UINT16_MAX * 2){
+        QDBOCR *rocrs = tcmalloc(sizeof(*rocrs) * onum);
+        uint32_t counts[UINT16_MAX+1];
+        memset(counts, 0, sizeof(counts));
+        for(int i = 0; i < onum; i++){
+          counts[ocrs[i].hash]++;
+        }
+        for(int i = 0; i < UINT16_MAX; i++){
+          counts[i+1] += counts[i];
+        }
+        for(int i = onum - 1; i >= 0; i--){
+          rocrs[--counts[ocrs[i].hash]] = ocrs[i];
+        }
+        for(int i = 0; i < UINT16_MAX; i++){
+          int num = counts[i+1] - counts[i];
+          if(num > 1) qsort(rocrs + counts[i], num, sizeof(*rocrs),
+                            (int (*)(const void *, const void *))tccmpocrs);
+        }
+        int num = onum - counts[UINT16_MAX];
+        if(num > 1) qsort(rocrs + counts[UINT16_MAX], num, sizeof(*rocrs),
+                          (int (*)(const void *, const void *))tccmpocrs);
+        tcfree(ocrs);
+        ocrs = rocrs;
+      } else if(onum > 1){
+        qsort(ocrs, onum, sizeof(*ocrs),
+              (int (*)(const void *, const void *))tccmpocrs);
+      }
+    }
+    TCIDSET *idset = tcidsetnew(onum * QDBHJBNUMCO + 1);
+    res = tcmalloc(sizeof(*res) * onum + 1);
+    int rnum = 0;
+    int rem = (seq - 1) * 2;
+    int ri = 0;
+    while(ri < onum){
+      QDBOCR *ocr = ocrs + ri;
+      ri++;
+      if(ocr->seq > 0) continue;
+      int64_t id = ocr->id;
+      int32_t max = ocr->off;
+      uint16_t seq = 1;
+      for(int i = ri; i < onum; i++){
+        QDBOCR *tocr = ocrs + i;
+        if(tocr->id != id) break;
+        if(tocr->seq == seq && tocr->off == max + 2){
+          max = tocr->off;
+          seq++;
+        }
+      }
+      if(max == ocr->off + rem){
+        if(!tcidsetcheck(idset, id)){
+          res[rnum++] = id;
+          tcidsetmark(idset, id);
+        }
+        while(ri < onum && ocrs[ri].id == id){
+          ri++;
         }
       }
     }
-    qsort(ocrs, onum, sizeof(*ocrs), (int (*)(const void *, const void *))tccmpocrs);
-    tclistdel(keys);
+    tcidsetdel(idset);
+    tcfree(ocrs);
+    *np = rnum;
   } else {
-    TCIDSET *lidset = NULL;
-    for(int i = 0; i < anum; i++){
-      int diff;
-      if(i & 0x1){
-        if(i != anum - 2) continue;
-        diff = 1;
-      } else {
-        if(i == anum - 1) continue;
-        diff = 2;
-      }
-      tcstrucstoutf(ary + i, 2, token);
-      int tlen = strlen(token);
-      int vsiz;
-      const char *vbuf = tcbdbget3(bdb, token, tlen, &vsiz);
-      if(!vbuf){
-        onum = 0;
-        continue;
-      }
-      int base = onum;
-      while(vsiz > 1){
+    TCIDSET *idset = tcidsetnew(QDBBITMAPNUM / 8 + 1);
+    res = tcmalloc(sizeof(*res) * QDBOCRUNIT);
+    int ranum = QDBOCRUNIT;
+    int rnum = 0;
+    BDBCUR *cur = tcbdbcurnew(idx);
+    tcbdbcurjump(cur, word, wsiz);
+    TCXSTR *key = tcxstrnew();
+    TCXSTR *val = tcxstrnew();
+    bool pchk = smode == QDBSPREFIX || smode == QDBSFULL;
+    bool schk = smode == QDBSSUFFIX || smode == QDBSFULL;
+    for(int i = 0; i < qdb->fwmmax && tcbdbcurrec(cur, key, val); i++){
+      const char *kbuf = TCXSTRPTR(key);
+      int ksiz = TCXSTRSIZE(key);
+      if(ksiz < wsiz || memcmp(kbuf, word, wsiz)) break;
+      if(schk && (ksiz != wsiz || memcmp(kbuf, word, wsiz))) break;
+      const char *cbuf = TCXSTRPTR(val);
+      int csiz = TCXSTRSIZE(val);
+      while(csiz > 0){
+        int64_t id = 0;
         int step;
-        uint64_t id;
-        TDREADVNUMBUF64(vbuf, id, step);
-        vbuf += step;
-        vsiz -= step;
-        if(vsiz > 0){
-          uint32_t off;
-          TDREADVNUMBUF(vbuf, off, step);
-          vbuf += step;
-          vsiz -= step;
-          if(onum >= oanum){
-            oanum *= 2;
-            ocrs = tcrealloc(ocrs, oanum * sizeof(*ocrs));
-          }
-          if(!lidset || tcidsetcheck(lidset, id)){
-            ocrs[onum].off = off | INT32_MIN;
-            ocrs[onum].id = id;
-            onum++;
+        TDREADVNUMBUF64(cbuf, id, step);
+        cbuf += step;
+        csiz -= step;
+        if(csiz > 0){
+          int off;
+          TDREADVNUMBUF(cbuf, off, step);
+          cbuf += step;
+          csiz -= step;
+          if((!pchk || off == 0) && !tcidsetcheck(idset, id)){
+            if(rnum >= ranum){
+              ranum *= 2;
+              res = tcrealloc(res, sizeof(*res) * ranum);
+            }
+            res[rnum++] = id;
+            tcidsetmark(idset, id);
           }
         }
       }
-      if(i == 0){
-        if(smode == QDBSPREFIX || smode == QDBSFULL){
-          int nnum = 0;
-          for(int j = 0; j < onum; j++){
-            if(ocrs[j].off == INT32_MIN){
-              ocrs[nnum].id = ocrs[j].id;
-              ocrs[nnum].off = 0;
-              nnum++;
-            }
-          }
-          onum = nnum;
-        } else {
-          for(int j = 0; j < onum; j++){
-            ocrs[j].off &= INT32_MAX;
-          }
-        }
-        if(anum == 2 && smode != QDBSFULL)
-          qsort(ocrs, onum, sizeof(*ocrs), (int (*)(const void *, const void *))tccmpocrs);
-      } else {
-        if(i <= 2){
-          TCIDSET *tidset = tcidsetnew((onum - base) * QDBHJBNUMCO + 1);
-          for(int j = base; j < onum; j++){
-            tcidsetmark(tidset, ocrs[j].id);
-          }
-          int nnum = 0;
-          for(int j = 0; j < base; j++){
-            if(tcidsetcheck(tidset, ocrs[j].id)){
-              ocrs[nnum].id = ocrs[j].id;
-              ocrs[nnum].off = ocrs[j].off;
-              nnum++;
-            }
-          }
-          for(int j = base; j < onum; j++){
-            ocrs[nnum].id = ocrs[j].id;
-            ocrs[nnum].off = ocrs[j].off;
-            nnum++;
-          }
-          onum = nnum;
-          tcidsetdel(tidset);
-        }
-        qsort(ocrs, onum, sizeof(*ocrs), (int (*)(const void *, const void *))tccmpocrs);
-        int nnum = 0;
-        uint64_t lid = UINT64_MAX;
-        for(int j = 0; j < onum; j++){
-          uint64_t id = ocrs[j].id;
-          if(id == lid && ocrs[j].off > INT32_MIN){
-            uint32_t off = (ocrs[j].off & INT32_MAX) - diff;
-            if(off < INT32_MAX){
-              for(int k = j - 1; k >= 0 && ocrs[k].id == id; k--){
-                if(ocrs[k].off == off){
-                  ocrs[nnum].id = id;
-                  ocrs[nnum].off = ocrs[j].off & INT32_MAX;
-                  nnum++;
-                  break;
-                }
-              }
-            }
-          }
-          lid = id;
-        }
-        onum = nnum;
-      }
-      if(i < anum - 2 || smode == QDBSSUFFIX || smode == QDBSFULL){
-        if(lidset) tcidsetdel(lidset);
-        lidset = tcidsetnew(onum * QDBHJBNUMCO + 1);
-        for(int j = 0; j < onum; j++){
-          tcidsetmark(lidset, ocrs[j].id);
-        }
-      }
-      if(onum < 1) break;
+      tcbdbcurnext(cur);
     }
-    if(smode == QDBSSUFFIX || smode == QDBSFULL){
-      tcstrucstoutf(ary + (anum - 1), 2, token);
-      int tlen = strlen(token);
-      int vsiz;
-      const char *vbuf = tcbdbget3(bdb, token, tlen, &vsiz);
-      while(vsiz > 1){
-        int step;
-        uint64_t id;
-        TDREADVNUMBUF64(vbuf, id, step);
-        vbuf += step;
-        vsiz -= step;
-        if(vsiz > 0){
-          uint32_t off;
-          TDREADVNUMBUF(vbuf, off, step);
-          vbuf += step;
-          vsiz -= step;
-          if(onum >= oanum){
-            oanum *= 2;
-            ocrs = tcrealloc(ocrs, oanum * sizeof(*ocrs));
-          }
-          if(!lidset || tcidsetcheck(lidset, id)){
-            ocrs[onum].off = off;
-            ocrs[onum].id = id;
-            onum++;
-          }
-        }
-      }
-      qsort(ocrs, onum, sizeof(*ocrs), (int (*)(const void *, const void *))tccmpocrs);
-      int nnum = 0;
-      QDBOCR locr;
-      locr.id = UINT64_MAX;
-      locr.off = UINT32_MAX;
-      for(int j = 0; j < onum; j++){
-        if(ocrs[j].id == locr.id && ocrs[j].off == locr.off + 1){
-          ocrs[nnum].id = ocrs[j].id;
-          ocrs[nnum].off = ocrs[j].off;
-          nnum++;
-        }
-        locr = ocrs[j];
-      }
-      onum = nnum;
-    }
-    if(lidset) tcidsetdel(lidset);
+    tcxstrdel(val);
+    tcxstrdel(key);
+    tcbdbcurdel(cur);
+    tcidsetdel(idset);
+    qsort(res, rnum, sizeof(*res), (int (*)(const void *, const void *))tccmpuint64);
+    *np = rnum;
   }
-  uint64_t *res = (uint64_t *)ocrs;
-  int rnum = 0;
-  uint64_t lid = UINT64_MAX;
-  for(int i = 0; i < onum; i++){
-    uint64_t id = ocrs[i].id;
-    if(lid == id) continue;
-    res[rnum++] = id;
-    lid = id;
-  }
-  *np = rnum;
+  tcfree(ary);
   return res;
 }
 
@@ -1574,11 +1622,11 @@ static uint64_t *tcqdbsearchimpl(TCQDB *qdb, const char *word, int smode, int *n
    `b' specifies the pointer to the other element.
    The return value is positive if the former is big, negative if the latter is big, 0 if both
    are equivalent. */
-static int tccmptokens(const TCLISTDATUM *a, const TCLISTDATUM *b){
+static int tccmptokens(const uint16_t **a, const uint16_t **b){
   assert(a && b);
-  uint16_t *ao = (uint16_t *)a->ptr;
-  uint16_t *bo = (uint16_t *)b->ptr;
-  return ((ao[0] << 16L) + ao[1]) - ((bo[0] << 16L) + bo[1]);
+  uint32_t anum = ((*a)[0] << 16UL) + (*a)[1];
+  uint32_t bnum = ((*b)[0] << 16UL) + (*b)[1];
+  return (anum < bnum) ? -1 : anum > bnum;
 }
 
 
@@ -1589,8 +1637,8 @@ static int tccmptokens(const TCLISTDATUM *a, const TCLISTDATUM *b){
    are equivalent. */
 static int tccmpocrs(QDBOCR *a, QDBOCR *b){
   assert(a && b);
-  if(a->id != b->id) return !((a->id - b->id) >> 63);
-  if(a->off < b->off) return -1;
+  if(a->id > b->id) return 1;
+  if(a->id < b->id) return -1;
   return a->off - b->off;
 }
 
@@ -1602,7 +1650,7 @@ static int tccmpocrs(QDBOCR *a, QDBOCR *b){
    are equivalent. */
 static int tccmpuint64(const uint64_t *a, const uint64_t *b){
   assert(a && b);
-  return !((*a - *b) >> 63);
+  return (*a < *b) ? -1 : *a > *b;
 }
 
 
