@@ -1,6 +1,6 @@
 /*************************************************************************************************
  * The command line utility of the q-gram database API
- *                                                      Copyright (C) 2007-2008 Mikio Hirabayashi
+ *                                                               Copyright (C) 2007-2010 FAL Labs
  * This file is part of Tokyo Dystopia.
  * Tokyo Dystopia is free software; you can redistribute it and/or modify it under the terms of
  * the GNU Lesser General Public License as published by the Free Software Foundation; either
@@ -54,7 +54,8 @@ static int procout(const char *path, int64_t id, const char *text, int omode, in
 static int procsearch(const char *path, const char **words, int wnum, int omode,
                       int topts, int emode, int smode, int max, bool ph);
 static int procoptimize(const char *path, int omode);
-static int procimporttsv(const char *path, const char *file, int64_t icsiz, int omode, int topts);
+static int procimporttsv(const char *path, const char *file,
+                         int64_t icsiz, int omode, int topts);
 static int procnormalize(const char *text, int topts);
 static int procversion(void);
 
@@ -64,7 +65,7 @@ int main(int argc, char **argv){
   g_progname = argv[0];
   g_dbgfd = -1;
   const char *ebuf = getenv("TCDBGFD");
-  if(ebuf) g_dbgfd = atoi(ebuf);
+  if(ebuf) g_dbgfd = tcatoi(ebuf);
   if(argc < 2) usage();
   int rv = 0;
   if(!strcmp(argv[1], "create")){
@@ -97,7 +98,7 @@ static void usage(void){
   fprintf(stderr, "%s: the command line utility of the q-gram database API\n", g_progname);
   fprintf(stderr, "\n");
   fprintf(stderr, "usage:\n");
-  fprintf(stderr, "  %s create [-tl] [-td|-tb] path [etnum]\n", g_progname);
+  fprintf(stderr, "  %s create [-tl] [-td|-tb|-tt] path [etnum]\n", g_progname);
   fprintf(stderr, "  %s inform [-nl|-nb] path\n", g_progname);
   fprintf(stderr, "  %s put [-nl|-nb] [-rc] [-ra] [-rs] path id text\n", g_progname);
   fprintf(stderr, "  %s out [-nl|-nb] [-rc] [-ra] [-rs] path id text\n", g_progname);
@@ -123,19 +124,26 @@ static void printerr(TCQDB *qdb){
 
 /* read a line from a file descriptor */
 static char *mygetline(FILE *ifp){
-  char *buf;
-  int c, len, blen;
-  buf = NULL;
-  len = 0;
-  blen = 256;
+  int len = 0;
+  int blen = 1024;
+  char *buf = tcmalloc(blen);
+  bool end = true;
+  int c;
   while((c = fgetc(ifp)) != EOF){
-    if(blen <= len) blen *= 2;
-    buf = tcrealloc(buf, blen + 1);
+    end = false;
+    if(c == '\0') continue;
+    if(blen <= len){
+      blen *= 2;
+      buf = tcrealloc(buf, blen + 1);
+    }
     if(c == '\n' || c == '\r') c = '\0';
     buf[len++] = c;
     if(c == '\0') break;
   }
-  if(!buf) return NULL;
+  if(end){
+    tcfree(buf);
+    return NULL;
+  }
   buf[len] = '\0';
   return buf;
 }
@@ -160,6 +168,8 @@ static int runcreate(int argc, char **argv){
       } else if(!strcmp(argv[i], "-td")){
         opts |= QDBTDEFLATE;
       } else if(!strcmp(argv[i], "-tb")){
+        opts |= QDBTBZIP;
+      } else if(!strcmp(argv[i], "-tt")){
         opts |= QDBTTCBS;
       } else {
         usage();
@@ -319,7 +329,7 @@ static int runsearch(int argc, char **argv){
         smode = QDBSFULL;
       } else if(!strcmp(argv[i], "-max")){
         if(++i >= argc) usage();
-        max = atoi(argv[i]);
+        max = tcatoi(argv[i]);
       } else if(!strcmp(argv[i], "-ph")){
         ph = true;
       } else {
@@ -374,15 +384,7 @@ static int runimporttsv(int argc, char **argv){
     if(!path && argv[i][0] == '-'){
       if(!strcmp(argv[i], "-ic")){
         if(++i >= argc) usage();
-        char *suffix;
-        icsiz = strtoll(argv[i], &suffix, 10);
-        if(*suffix == 'k' || *suffix == 'K'){
-          icsiz *= 1024LL;
-        } else if(*suffix == 'm' || *suffix == 'M'){
-          icsiz *= 1024LL * 1024LL;
-        } else if(*suffix == 'g' || *suffix == 'G'){
-          icsiz *= 1024LL * 1024LL * 1024LL;
-        }
+        icsiz = tcatoix(argv[i]);
       } else if(!strcmp(argv[i], "-nl")){
         omode |= HDBONOLCK;
       } else if(!strcmp(argv[i], "-nb")){
@@ -603,15 +605,15 @@ static int procsearch(const char *path, const char **words, int wnum, int omode,
     int rnum;
     uint64_t *res;
     switch(emode){
-    case EMUNION:
-      res = tcqdbresunion(rsets, wnum, &rnum);
-      break;
-    case EMDIFF:
-      res = tcqdbresdiff(rsets, wnum, &rnum);
-      break;
-    default:
-      res = tcqdbresisect(rsets, wnum, &rnum);
-      break;
+      case EMUNION:
+        res = tcqdbresunion(rsets, wnum, &rnum);
+        break;
+      case EMDIFF:
+        res = tcqdbresdiff(rsets, wnum, &rnum);
+        break;
+      default:
+        res = tcqdbresisect(rsets, wnum, &rnum);
+        break;
     }
     if(ph){
       printf("hits: %d\n", rnum);
@@ -660,7 +662,8 @@ static int procoptimize(const char *path, int omode){
 
 
 /* perform importtsv command */
-static int procimporttsv(const char *path, const char *file, int64_t icsiz, int omode, int topts){
+static int procimporttsv(const char *path, const char *file,
+                         int64_t icsiz, int omode, int topts){
   TCQDB *qdb = tcqdbnew();
   if(g_dbgfd >= 0) tcqdbsetdbgfd(qdb, g_dbgfd);
   FILE *ifp = file ? fopen(file, "rb") : stdin;
@@ -732,8 +735,9 @@ static int procnormalize(const char *text, int topts){
 
 /* perform version command */
 static int procversion(void){
-  printf("Tokyo Dystopia version %s (%d:%s)\n", tdversion, _TD_LIBVER, _TD_FORMATVER);
-  printf("Copyright (C) 2007-2008 Mikio Hirabayashi\n");
+  printf("Tokyo Dystopia version %s (%d:%s) for %s\n",
+         tdversion, _TD_LIBVER, _TD_FORMATVER, TDSYSNAME);
+  printf("Copyright (C) 2007-2010 FAL Labs\n");
   return 0;
 }
 
